#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import base64
import logging
import os
import random
import string
from datetime import datetime

import pytz

from ibmfl.crypto.crypto_exceptions import KeyDistributionInputException, KeyDistributionVerificationException
from ibmfl.crypto.crypto_library import Crypto
from ibmfl.crypto.infra.crypto_asym_imp_rsa import CryptoAsymRsa
from ibmfl.crypto.infra.crypto_cert_imp_rsa import CryptoCertRsa
from ibmfl.crypto.infra.crypto_he_imp_hely import CryptoHeLy
from ibmfl.crypto.infra.crypto_sym_imp_fernet import CryptoSymFernet

logger = logging.getLogger(__name__)


class CryptoKeysProtoParty:
    """
    This class implements the party side of the protocol for generating and distributing
    cryptographic keys among the parties via the aggregator.
    """

    def __init__(self, config_dst: dict, config_crypto: dict = None, crypto_sys: Crypto = None):
        """
        Initializes the party side protocol object.
        Loads the CA cert, my cert, and additional configuration information
        for the party to implement the keys generation and distribution protocol.

        :param config_dst: Configuration for the keys distribution protocol.
        :type config: `dict`
        :param config_crypto: Configuration for the crypto system.
        :type config: `dict`
        :param crypto_sys: Crypto system object that supports homomorphic encryption.
        :type crypto_sys: `Crypto`
        """

        self.id = None
        self.crypto_cert = CryptoCertRsa(config_dst["ca_cert_file_path"], config_dst["my_cert_file_path"])
        if "asym_key_password" in config_dst:
            asym_pass = config_dst["asym_key_password"]
            if isinstance(asym_pass, str):
                asym_pass = asym_pass.encode()
            elif not isinstance(asym_pass, bytes):
                raise KeyDistributionInputException("Invalid type of asym_pass=" + str(type(asym_pass)))
        else:
            asym_pass = None
        self.crypto_asym = CryptoAsymRsa(config_dst["asym_key_file_path"], asym_pass)
        self.crypto_sym = CryptoSymFernet()
        self.crypto_he = CryptoHeLy(config_crypto)
        self.he_pb = None
        self.he_pr = None
        self.keys_id = None
        self.keys_time = None
        self.crypto_sys = crypto_sys
        if "cp_certs_folder_path" in config_dst and isinstance(config_dst["cp_certs_folder_path"], str):
            self.cp_certs_folder_path = config_dst["cp_certs_folder_path"]
            self.cp_certs_str = {}
            for filename in os.scandir(self.cp_certs_folder_path):
                if filename.is_file():
                    id = os.path.basename(filename.path).split("_")[0]
                    with open(filename.path) as f:
                        self.cp_certs_str[id] = f.read()
        else:
            self.cp_certs_folder_path = None
            self.cp_certs_str = None
        return

    def get_he_keys(self) -> tuple:
        """
        Returns the HE key pair.
        """
        return (self.he_pb, self.he_pr)

    def get_my_cert(self, info: dict) -> dict:
        """
        Returns a payload including this party's ID and certificate.

        :param info: Dictionary that includes this party's ID.
        :type info: `dict`
        :return: Returns a payload including this party's ID and certificate.
        :rtype: `dict`
        """
        self.id = info["id"]
        logger.info("get_my_cert: [id: " + str(self.id) + "] [payload: " + str(info) + "]")
        msg = {"id": self.id, "cert": self.crypto_cert.get_my_cert("pem")}
        return msg

    def get_keys_id(self):
        """
        Returns the keys ID.
        """
        return self.keys_id

    def get_keys_time(self):
        """
        Returns the creation time of the keys.
        """
        return self.keys_time

    def __verify_parties_list(self, party_certs: dict, allowed_parties_info: dict):
        """
        Verifies that the parties in the provided list of parties' certificates are included in a
        verified list of allowed parties.

        :param party_certs: List of parties' certificates.
        :type party_certs: `dict`
        :param allowed_parties_info: Information on allowed parties.
        :type allowed_parties_info: `dict`
        """

        if self.cp_certs_str is None:
            raise KeyDistributionInputException("self.cp_certs_str is None")
        # Check that the received cp_cert is one of the configured cp_certs.
        cp_id = allowed_parties_info["cp_id"]
        if cp_id not in self.cp_certs_str:
            raise KeyDistributionInputException(
                "cp_id=" + str(cp_id) + " not in self.cp_certs_str=" + str(self.cp_certs_str.keys())
            )
        if allowed_parties_info["cp_cert"] != self.cp_certs_str[cp_id]:
            raise KeyDistributionInputException(
                "allowed_parties_info['cp_cert']="
                + str(allowed_parties_info["cp_cert"])
                + " != self.cp_certs_str[cp_id]="
                + str(self.cp_certs_str[cp_id])
            )
        # Verify the cp_cert.
        verf, cp_pbk = self.crypto_cert.verify_cert_signature(self.cp_certs_str[cp_id].encode())
        if not verf:
            raise KeyDistributionVerificationException("Invalid certificate=" + str(self.cp_certs_str[cp_id]))
        # Verify the cp signature on the list of IDs.
        signature = base64.b64decode(allowed_parties_info["cp_signature"].encode())
        verf = CryptoAsymRsa.verify_signature(cp_pbk, signature, allowed_parties_info["ids"].encode())
        if not verf:
            raise KeyDistributionVerificationException("Invalid signature=" + str(allowed_parties_info["cp_signature"]))
        # Verify that all the received IDs are included in the list of allowed IDs.
        allowed_pids = allowed_parties_info["ids"].split(",")
        for pid in party_certs.keys():
            if pid not in allowed_pids:
                raise KeyDistributionInputException(
                    "pid=" + str(pid) + " not in allowed_pids=" + str(allowed_parties_info["ids"])
                )
        return

    def generate_keys(self, info: dict) -> dict:
        """
        Generates keys and a payload of a keys distribution message.
        Outgoing message structure:
        {my_cert|he_pb|he_pb_sign|he_pr_enc|he_pr_enc_sign|keys_id|keys_time|{id->(sym_enc|sym_enc_sign)}}

        :param info: Dictionary containing party_certs (certificates of parties) and allowed_parties_info.
        :type info: `dict`
        :return: Payload of the generated keys distribution message.
        :rtype: `dict`
        """

        logger.info("generate_keys: [id: " + str(self.id) + "] start")
        logger.debug("generate_keys: [id: " + str(self.id) + "] [payload: " + str(info) + "]")

        # Verify input.
        if "party_certs" not in info or info["party_certs"] is None:
            raise KeyDistributionInputException("party_certs not in info or info['party_certs'] is None")
        party_certs = info["party_certs"]

        # Verify the parties list.
        if "allowed_parties_info" in info and info["allowed_parties_info"] is not None:
            self.__verify_parties_list(party_certs, info["allowed_parties_info"])

        # Verify the party certificates and extract their public keys.
        party_certs.pop(self.id, None)
        party_keys = self.crypto_cert.verify_certs(party_certs)

        # Generate keys and meta properties.
        self.__generate_keys_elements()
        logger.info("generate_keys: [id: " + str(self.id) + "] generated keys")

        # Set the HE keys in the crypto system object.
        if self.crypto_sys is not None:
            keys = {"pp": self.he_pb, "sk": self.he_pr}
            self.crypto_sys.set_keys(keys)

        # Generate a distribution message.
        msg = self.__generate_distribution_message(party_keys)

        logger.info("generate_keys: [id: " + str(self.id) + "] [len(party_keys): " + str(len(party_keys)) + "] end")
        return msg

    def distribute_keys(self, info: dict) -> dict:
        """
        Generates a keys distribution message to distribute the existing keys.
        Outgoing message structure:
        {my_cert|he_pb|he_pb_sign|he_pr_enc|he_pr_enc_sign|keys_id|keys_time|{id->(sym_enc|sym_enc_sign)}}

        :param info: Dictionary containing party_certs (certificates of parties) and allowed_parties_info.
        :type info: `dict`
        :return: Payload of the keys distribution message.
        :rtype: `dict`
        """

        logger.info("distribute_keys: [id: " + str(self.id) + "] start")
        logger.debug("distribute_keys: [id: " + str(self.id) + "] [payload: " + str(info) + "]")

        # Verify input.
        if "party_certs" not in info or info["party_certs"] is None:
            raise KeyDistributionInputException("party_certs not in info or info['party_certs'] is None")
        party_certs = info["party_certs"]

        # Verify the parties list.
        if "allowed_parties_info" in info and info["allowed_parties_info"] is not None:
            self.__verify_parties_list(party_certs, info["allowed_parties_info"])

        # Verify the party certificates and extract their public keys.
        party_certs.pop(self.id, None)
        if len(party_certs) > 0:
            party_keys = self.crypto_cert.verify_certs(party_certs)
        else:
            party_keys = None

        # Generate a distribution message.
        msg = self.__generate_distribution_message(party_keys)

        logger.info("distribute_keys: [id: " + str(self.id) + "] [len(party_keys): " + str(len(party_keys)) + "] end")
        return msg

    def __generate_distribution_message(self, party_keys: dict) -> dict:
        """
        Generates a keys distribution message in a dictionary form.

        :return: Keys distribution message in a dictionary form.
        :rtype: `dict`
        """

        msg = {}

        # Store meta properties in the message.
        msg["gen_cert"] = self.crypto_cert.get_my_cert("pem")
        msg["keys_id"] = self.keys_id
        msg["keys_time"] = self.keys_time

        # Store public keys in the message.
        msg["he_pb"] = self.he_pb
        msg["he_pb_sign"] = self.crypto_asym.get_signature(self.he_pb)

        # Store secret keys in the message.
        if party_keys is not None and len(party_keys) > 0:
            # Encrypted HE private key.
            he_pr_enc = self.crypto_sym.encrypt(self.he_pr)
            msg["he_pr_enc"] = he_pr_enc
            msg["he_pr_enc_sign"] = self.crypto_asym.get_signature(he_pr_enc)

            # Dict of sym key encrypted per party.
            party_sym = {}
            sym_bytes = self.crypto_sym.get_key()
            for id, pr_pbk in party_keys.items():
                sym_enc = CryptoAsymRsa.encrypt_wkey(pr_pbk, sym_bytes)
                sym_enc_sign = self.crypto_asym.get_signature(sym_enc)
                party_sym[id] = (sym_enc, sym_enc_sign)
            msg["party_sym"] = party_sym

        return msg

    def __generate_keys_elements(self):
        """
        Generates the various keys and meta properties and stores these in the object.
        """

        self.crypto_sym.generate_key()
        self.crypto_he.generate_keys()
        self.he_pb = self.crypto_he.get_public_key()
        self.he_pr = self.crypto_he.get_private_key()
        self.keys_id = generate_random_string(64)
        self.keys_time = datetime.now(pytz.timezone("America/New_York"))
        return

    def parse_keys(self, keys_msg: dict) -> bool:
        """
        Parses a keys distribution message.
        Incoming message structure:
        {gen_cert|he_pb|he_pb_sign|he_pr_enc|he_pr_enc_sign|sym_enc|sym_enc_sign|keys_id|keys_time}

        :param keys_msg: Keys distribution message payload.
        :type keys_msg: `dict`
        :return: Indication if the keys were parsed and set successfully.
        :rtype: `bool`
        """

        logger.info("parse_keys: [id: " + str(self.id) + "] start")
        logger.debug("parse_keys: [id: " + str(self.id) + "] [payload: " + str(keys_msg) + "]")

        # Verify the certificate of the generating party.
        ver, gen_pbk = self.crypto_cert.verify_cert_signature(keys_msg["gen_cert"])
        if not ver:
            raise KeyDistributionVerificationException(
                "Invalid generating party certificate=" + str(keys_msg["gen_cert"])
            )

        # Parse the sym key.
        sym_enc = keys_msg["sym_enc"]
        ver = CryptoAsymRsa.verify_signature(gen_pbk, keys_msg["sym_enc_sign"], sym_enc)
        if not ver:
            raise KeyDistributionVerificationException("Invalid signature on sym_enc=" + str(keys_msg["sym_enc_sign"]))
        sym_bytes = self.crypto_asym.decrypt(sym_enc)
        self.crypto_sym = CryptoSymFernet(sym_bytes)

        # Parse the he public key.
        self.he_pb = keys_msg["he_pb"]
        ver = CryptoAsymRsa.verify_signature(gen_pbk, keys_msg["he_pb_sign"], self.he_pb)
        if not ver:
            raise KeyDistributionVerificationException("Invalid signature on he_pb=" + str(keys_msg["he_pb_sign"]))

        # Parse the he private key.
        he_pr_enc = keys_msg["he_pr_enc"]
        ver = CryptoAsymRsa.verify_signature(gen_pbk, keys_msg["he_pr_enc_sign"], he_pr_enc)
        if not ver:
            raise KeyDistributionVerificationException(
                "Invalid signature on he_pr_enc=" + str(keys_msg["he_pr_enc_sign"])
            )
        self.he_pr = self.crypto_sym.decrypt(he_pr_enc)

        # Parse the keys ID and time.
        self.keys_id = keys_msg["keys_id"]
        self.keys_time = keys_msg["keys_time"]

        # Set the HE keys in the crypto system object.
        if self.crypto_sys is not None:
            keys = {"pp": self.he_pb, "sk": self.he_pr}
            self.crypto_sys.set_keys(keys)

        logger.info("parse_keys: [id: " + str(self.id) + "] end")
        return True


def generate_random_string(length) -> str:
    letters = string.ascii_lowercase
    result_str = "".join(random.choice(letters) for i in range(length))
    return result_str



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import mlpipelinepy


class MLPipelineVersion(type):
    _version = mlpipelinepy.__version__
    _version_parts = _version.split(".")

    @classmethod
    def major_ver(mcs):
        return int(mcs._version_parts[0])

    @classmethod
    def minor_ver(mcs):
        return int(mcs._version_parts[1])

    @classmethod
    def full_version(mcs):
        return mcs._version



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from .content_reader import ContentReader
from .ml_repository_api import MLRepositoryApi
from .ml_repository_client import MLRepositoryClient
from .model_adapter import ModelAdapter
from .model_collection import ModelCollection
from .experiment_adapter import ExperimentAdapter
from .experiment_collection import ExperimentCollection
from .ml_repository_client import connect
from .wml_experiment_collection import WmlExperimentCollection
from .wml_experiment_adapter import WmlExperimentCollectionAdapter
from .libraries_adapter import WmlLibrariesAdapter
from .libraries_collection import LibrariesCollection
from .runtimes_adapter import WmlRuntimesAdapter
from .runtimes_collection import RuntimesCollection


__all__ = ['ContentReader', 'MLRepositoryApi', 'MLRepositoryClient', 'ModelAdapter', 'ModelCollection',
           'ExperimentAdapter', 'ExperimentCollection', 'connect', 'WmlExperimentCollection', 'WmlExperimentCollectionAdapter', 'WmlLibrariesAdapter', 'LibrariesCollection',
           'RuntimesCollection', 'WmlRuntimesAdapter']



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from .artifact import Artifact


class ModelArtifact(Artifact):
    """
    Class representing model artifact.

    :param str uid: optional, uid which indicate that artifact already exists in repository service
    :param str name: optional, name of artifact
    :param MetaProps meta_props: optional, props used by other services
    """
    def __init__(self, uid, name, meta_props):
        super(ModelArtifact, self).__init__(uid, name, meta_props)

    def pipeline_artifact(self):
        """
        Returns pipeline artifact from which this model artifact was created.

        :rtype: ModelArtifact
        """
        try:
            return self._pipeline_artifact
        except:
            self._pipeline_artifact = self.client.pipelines.version_from_url(self._training_definition_version_url)
            return self._pipeline_artifact



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

class MetaPropMissingError(Exception):
    pass


class UnsupportedTFSerializationFormat(Exception):
    pass


class UnmatchedKerasVersion(Exception):
    pass

class InvalidCaffeModelArchive(Exception):
    pass


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.libs.repo.util.unique_id_gen import uid_generate
import os, shutil

class FunctionArtifactLoader(object):

    def load(self,path=None):
        return self.extract_content(path)

    def extract_content(self, file_path=None):
        if file_path is None:
            file_path = 'function_artifact'
            try:
                shutil.rmtree(file_path)
            except:
                pass
            os.makedirs(file_path)

        try:
            id_length = 20
            lib_generated_id = uid_generate(id_length)

            file_name = '{}/{}_{}.zip'.format(file_path, "function_artifact", lib_generated_id)
            input_stream = self.reader().read()
            file_content = input_stream.read()
            gz_f = open(file_name, 'wb+')
            gz_f.write(file_content)
            gz_f.close()
            return os.path.abspath(file_name)
        except Exception as ex:
            raise ex



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from abc import ABC, abstractmethod


class Serializer(ABC):
    """
    Abstract class for Serializer
    """
    @abstractmethod
    def serialize(self, msg):
        """
        Serialize a message

        :param msg: message to serialize
        :type msg: `Message`
        :return: serialized byte stream
        :rtype: `b[]`
        """
        pass

    @abstractmethod
    def deserialize(self, serialized_byte_stream):
        """
        Deserialize a byte stream to a message

        :param serialized_byte_stream: byte stream
        :type serialized_byte_stream: `b[]`
        :return: deserialized message
        :rtype: `Message`
        """
        pass



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import absolute_import

# import models into model package
from .array_model_metrics_output import ArrayModelMetricsOutput
from .array_model_output import ArrayModelOutput
from .array_model_version_output import ArrayModelVersionOutput
from .array_pipeline_output import ArrayPipelineOutput
from .array_pipeline_version_output import ArrayPipelineVersionOutput
from .artifact_author import ArtifactAuthor
from .artifact_version_metadata import ArtifactVersionMetadata
from .artifact_version_short_metadata import ArtifactVersionShortMetadata
from .batch_deploy_output import BatchDeployOutput
from .batch_deploy_output_entity import BatchDeployOutputEntity
from .batch_deploy_output_entity_execution import BatchDeployOutputEntityExecution
from .batch_deploy_output_meta import BatchDeployOutputMeta
from .batch_input import BatchInput
from .batch_output import BatchOutput
from .batch_output_array import BatchOutputArray
from .cols_output import ColsOutput
from .connection import Connection
from .deploy_input import DeployInput
from .error_message import ErrorMessage
from .evaluation_definition import EvaluationDefinition
from .evaluation_definition_metrics import EvaluationDefinitionMetrics
from .input_data_schema import InputDataSchema
from .internal_input_batch import InternalInputBatch
from .internal_output_batch import InternalOutputBatch
from .json_patch_array import JsonPatchArray
from .json_patch_entity import JsonPatchEntity
from .meta_object import MetaObject
from .meta_object_metadata import MetaObjectMetadata
from .model_input import ModelInput
from .model_metrics import ModelMetrics
from .model_metrics_values import ModelMetricsValues
from .model_output import ModelOutput
from .model_output_entity import ModelOutputEntity
from .model_output_entity_pipeline_version import ModelOutputEntityPipelineVersion
from .model_training_data_ref import ModelTrainingDataRef
from .model_type import ModelType
from .model_version_input import ModelVersionInput
from .model_version_output import ModelVersionOutput
from .model_version_output_entity import ModelVersionOutputEntity
from .model_version_output_entity_model import ModelVersionOutputEntityModel
from .pipeline_input import PipelineInput
from .pipeline_output import PipelineOutput
from .pipeline_output_entity import PipelineOutputEntity
from .pipeline_type import PipelineType
from .pipeline_version_input import PipelineVersionInput
from .pipeline_version_output import PipelineVersionOutput
from .pipeline_version_output_entity import PipelineVersionOutputEntity
from .pipeline_version_output_entity_parent import PipelineVersionOutputEntityParent
from .runtime_environment import RuntimeEnvironment
from .score_input import ScoreInput
from .score_output import ScoreOutput
from .spark_service import SparkService
from .stream_input_internal import StreamInputInternal
from .stream_internal import StreamInternal
from .stream_output import StreamOutput
from .stream_output_array import StreamOutputArray
from .stream_output_internal import StreamOutputInternal
from .token_response import TokenResponse
from .training_data_schema import TrainingDataSchema
from .content_location import ContentLocation
from .content_status import ContentStatus
from .output_data_schema import OutputDataSchema

# import v3 models into model package
from .array_data_input_repository import ArrayDataInputRepository
from .artifact_version_repository import ArtifactVersionRepository
from .author_repository import AuthorRepository
from .connection_object_with_name_repository import ConnectionObjectWithNameRepository
from .error_repository import ErrorRepository
from .error_repository_target import ErrorRepositoryTarget
from .error_schema_repository import ErrorSchemaRepository
from .evaluation_definition_repository import EvaluationDefinitionRepository
from .evaluation_definition_repository_metrics import EvaluationDefinitionRepositoryMetrics
from .framework_output_repository import FrameworkOutputRepository
from .framework_output_repository_runtimes import FrameworkOutputRepositoryRuntimes
from .framework_output_repository_libraries import  FrameworkOutputRepositoryLibraries
from .meta_object_repository import MetaObjectRepository
from .meta_object_repository_metadata import MetaObjectRepositoryMetadata
from .ml_assets_create_experiment_input import MlAssetsCreateExperimentInput
from .ml_assets_create_experiment_output import MlAssetsCreateExperimentOutput
from .ml_assets_create_experiment_output_array import MlAssetsCreateExperimentOutputArray
from .ml_assets_create_model_input import MlAssetsCreateModelInput
from .ml_assets_create_model_output import MlAssetsCreateModelOutput
from .ml_assets_create_model_output_array import MlAssetsCreateModelOutputArray
from .ml_assets_model_size_output import MlAssetsModelSizeOutput
from .ml_assets_get_presigned_url_output import MlAssetsGetPreSignedUrlOutput
from .ml_assets_upload_content_output import MlAssetsUploadContentOutput
from .ml_assets_upload_content_output_metadata import MlAssetsUploadContentOutputMetadata
from .ml_assets_upload_content_output_entity import MlAssetsUploadContentOutputEntity


from .model_content_location import ModelContentLocation
from .training_models import TrainingModels
from .space_models import SpaceModels
from .runtime_models import RuntimeModels
from .software_spec_models import SoftwareSpecModels
from .model_definition_models import ModelDefinitionModels
from .model_schemas import ModelSchemas
from .pipeline_models import PipelinesModels
from .model_content_location import ModelContentLocation
from .custom_models import ModelsCustom
from .metrics_models import ModelsMetrics
from .model_schemas import ModelSchemas
from .size_models import ModelsSize
from .training_models import TrainingModels








# import experiments
from .array_model_version_metrics_experiments import ArrayModelVersionMetricsExperiments
from .array_training_output_experiments import ArrayTrainingOutputExperiments
from .author_experiments import AuthorExperiments
from .compute_configuration_experiments import ComputeConfigurationExperiments
from .connection_object_source_experiments import ConnectionObjectSourceExperiments
from .connection_object_target_experiments import ConnectionObjectTargetExperiments
from .error_experiments import ErrorExperiments
from .error_experiments_target import ErrorExperimentsTarget
from .error_schema_experiments import ErrorSchemaExperiments
from .evaluation_definition_experiments import EvaluationDefinitionExperiments
from .tag_repository import TagRepository
from .experiment_input import ExperimentInput
from .experiment_input_settings import ExperimentInputSettings
from .experiment_output_array import ExperimentOutputArray
from .experiment_output_array_first import ExperimentOutputArrayFirst
from .experiment_output import ExperimentOutput
from .experiment_patch import ExperimentPatch
from .experiment_status_experiments import ExperimentStatusExperiments
from .hyper_parameters_experiments import HyperParametersExperiments
from .hyper_parameters_experiments_inner_values_range import HyperParametersExperimentsInnerValuesRange
from .hyper_parameters_for_status_experiments import HyperParametersForStatusExperiments
from .hyper_parameters_for_status_experiments_inner import HyperParametersForStatusExperimentsInner
from .hyper_parameters_optimization_experiments import HyperParametersOptimizationExperiments
from .meta_object_experiments import MetaObjectExperiments
from .meta_object_experiments_metadata import MetaObjectExperimentsMetadata
from .metric_object_experiments import MetricObjectExperiments
from .model_version_metrics_experiments import ModelVersionMetricsExperiments
from .patch_operation_experiments import PatchOperationExperiments
from .training_output_experiments import TrainingOutputExperiments
from .training_reference_experiments import TrainingReferenceExperiments
from .training_status_experiments import TrainingStatusExperiments
from .training_status_experiments_result import TrainingStatusExperimentsResult
from .hyper_parameters import HyperParameters
from .hyper_parameters_optimization_experiments_method import HyperParametersOptimizationExperimentsMethod
from .hyper_parameters_optimization_experiments_method_parameters import HyperParametersOptimizationExperimentsMethodParameters
from .hyper_parameters_experiments_double_range import HyperParametersExperimentsDoubleRange
from .hyper_parameters_experiments_int_range import HyperParametersExperimentsIntRange
from .libraries_definition_input import LibrariesDefinitionInput
from .patch_operation_libraries import PatchOperationLibraries
from .ml_assets_patch_libraries_input import MlAssetsPatchLibrariesInput
from .ml_assets_create_libraries_output import MlAssetsCreateLibrariesOutput
from .ml_assets_create_libraries_output_array_first import MlAssetsCreateLibrariesOutputArrayFirst
from .ml_assets_create_libraries_output_array import MlAssetsCreateLibrariesOutputArray
from .ml_assets_create_patch_libraries_output import MlAssetsCreatePatchLibrariesOutput
from .runtime_output_repository import RuntimeOutputRepository
from .runtime_spec_definition_input import RuntimeSpecDefinitionInput
from .runtime_spec_definition_input_platform import RuntimeSpecDefinitionInputPlatform
from .runtime_spec_definition_input_public_libraries import RuntimeSpecDefinitionInputPublicLibraries
from .runtime_spec_definition_input_repository import RuntimeSpecDefinitionInputRepository
from .ml_assets_create_runtime_spec_output import MlAssetsCreateRuntimeSpecOutput
from .ml_assets_create_runtime_spec_output_array import MlAssetsCreateRuntimeSpecOutputArray
from .ml_assets_create_runtime_spec_output_array_first import MlAssetsCreateRuntimeSpecOutputArrayFirst
from .ml_assets_patch_runtime_spec_input import MlAssetsPatchRuntimeSpecInput
from .ml_assets_patch_runtime_spec_output import MlAssetsPatchRuntimeSpecOutput
from .patch_operation_runtime_spec import PatchOperationRuntimeSpec
from .runtime_spec_definition_input_custom_libraries import RuntimeSpecDefinitionInputCustomLibraries
from .libraries_definition_input_platform import LibrariesDefinitionInputPlatform
from .model_definition_models import ModelDefinitionModels




from .ml_assets_create_function_input import MlAssetsCreateFunctionInput
from .ml_assets_create_functions_output import MlAssetsCreateFunctionsOutput
from .meta_object_functions_metadata import MetaObjectFunctionsMetadata
from .ml_assets_create_functions_output_array import MlAssetsCreateFunctionsOutputArray
from .ml_assets_create_functions_output_array_first import MlAssetsCreateFunctionsOutputArrayFirst
from .patch_operation_functions import PatchOperationFunctions





# coding: utf-8

"""
    IBM Watson Machine Learning REST API

    ## Authorization  ### IBM Watson Machine Learning Credentials  To start working with API one needs to generate an `access token` using the `username` and `password` available on the Service Credentials tab of the IBM Watson Machine Learning service instance or also available in the VCAP environment variable.  Example of the Service Credentials:  ```json {     \"url\": \"https://ibm-watson-ml.mybluemix.net\",     \"access_key\": \"ERY9vcBfE4sE+F4g8hcotF9L+j81WXWeZv\",     \"username\": \"c1ef4b80-2ee2-458e-ab92-e9ca97ec657d\",     \"password\": \"030528d4-5a3e-4d4c-9258-5d553513be6f\" } ```  Example of obtaining `access token` from Token Endpoint using HTTP Basic Auth (for details please refer to Token section below):  ` curl --basic --user username:password https://ibm-watson-ml.mybluemix.net/v2/identity/token `   ### Apache Spark Service Credentials  The IBM Watson Machine Learning co-operates with the Apache Spark as a Service to deploy pipeline models. For API methods requiring Apache Spark Service instance a custom header: `X-Spark-Service-Instance` with Service Credentials must be specified. The header value is a **base64 encoded** string with the Service Credentials JSON data.  [Example of API method requiring Apache Spark Service](https://console.ng.bluemix.net/docs/services/PredictiveModeling/index-gentopic1.html#pm_service_api_spark_batch) 

    OpenAPI spec version: 2.0.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class ArrayPipelineOutput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, resources=None):
        """
        ArrayPipelineOutput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'resources': 'list[PipelineOutput]'
        }

        self.attribute_map = {
            'resources': 'resources'
        }

        self._resources = resources

    @property
    def resources(self):
        """
        Gets the resources of this ArrayPipelineOutput.


        :return: The resources of this ArrayPipelineOutput.
        :rtype: list[PipelineOutput]
        """
        return self._resources

    @resources.setter
    def resources(self, resources):
        """
        Sets the resources of this ArrayPipelineOutput.


        :param resources: The resources of this ArrayPipelineOutput.
        :type: list[PipelineOutput]
        """

        self._resources = resources

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import abc
import logging

from ibmfl.data.data_handler import DataHandler

logger = logging.getLogger(__name__)


class PandasDataHandler(DataHandler):
    """
    Base class to load and pre-process data.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.x_train = None
        self.y_train = None
        self.x_test = None
        self.y_test = None

    @abc.abstractmethod
    def get_data(self):
        """
        Read data and return as Pandas data frame.

        :return: A dataset structure
        :rtype: `pandas.core.frame.DataFrame`
        """

    @abc.abstractmethod
    def get_dataset_info(self, **kwargs):
        """
        Read and extract data information

        :return: some information about the dataset (i.e. a dictionary that contains the list of features)
        :rtype: `dict`
        """
        raise NotImplemented

    def get_min(self, dp_flag=False, **kwargs):
        """
        Assuming the dataset is loaded as type `pandas.DataFrame`, and
        has shape(num_samples, num_features).

        :param dp_flag: Flag for differential private answer. By default is \
        set to False.
        :type dp_flag: `boolean`
        :param kwargs: Dictionary of differential privacy arguments \
        for computing the minimum value of each feature across all samples, \
        e.g., epsilon and delta, etc.
        :type kwargs: `dict`
        :return: A vector of shape (1, num_features) stores the minimum value \
        of each feature across all samples.
        :rtype: `pandas.Series` where each entry matches the original type \
        of the corresponding feature.
        """
        train_data, (_) = self.get_data()

        if not dp_flag:
            logger.info('Calculating minimum values.')
            min_vec = train_data.min(axis=0)
            # TODO dp minimum calculation
        return min_vec



"""
Module to where data handler are implemented.
"""
#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import abc

from ibmfl.data.data_handler import DataHandler
from ibmfl.data.env_spec import EnvHandler


class EnvDataHandler(DataHandler):
    """
    Base class to load data and  environment for reinforcement learning.
    """

    @abc.abstractmethod
    def get_data(self, **kwargs):
        """
        Read train data and test data for reinforcement learning
        :return:
        """

    @abc.abstractmethod
    def get_env_class_ref(self) -> EnvHandler:
        """
           Get environment reference for RL trainer, the instance is created in
           model class as part of trainer initialization
        """



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class ModelInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None, description=None, author=None, pipeline_version_href=None, type=None, runtime_environment=None, training_data_schema=None, label_col=None, input_data_schema=None, output_data_schema=None):
        """
        ModelInput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'description': 'str',
            'author': 'ArtifactAuthor',
            'pipeline_version_href': 'str',
            'type': 'ModelType',
            'runtime_environment': 'RuntimeEnvironment',
            'training_data_schema': 'object',
            'label_col': 'str',
            'input_data_schema': 'object',
            'output_data_schema': 'object'
        }

        self.attribute_map = {
            'name': 'name',
            'description': 'description',
            'author': 'author',
            'pipeline_version_href': 'pipelineVersionHref',
            'type': 'type',
            'runtime_environment': 'runtimeEnvironment',
            'training_data_schema': 'trainingDataSchema',
            'label_col': 'labelCol',
            'input_data_schema': 'inputDataSchema',
            'output_data_schema': 'outputDataSchema'
        }

        self._name = name
        self._description = description
        self._author = author
        self._pipeline_version_href = pipeline_version_href
        self._type = type
        self._runtime_environment = runtime_environment
        self._training_data_schema = training_data_schema
        self._label_col = label_col
        self._input_data_schema = input_data_schema
        self._output_data_schema = output_data_schema

    @property
    def name(self):
        """
        Gets the name of this ModelInput.
        Name of the model

        :return: The name of this ModelInput.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this ModelInput.
        Name of the model

        :param name: The name of this ModelInput.
        :type: str
        """

        self._name = name

    @property
    def description(self):
        """
        Gets the description of this ModelInput.
        Description of the model

        :return: The description of this ModelInput.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this ModelInput.
        Description of the model

        :param description: The description of this ModelInput.
        :type: str
        """

        self._description = description

    @property
    def author(self):
        """
        Gets the author of this ModelInput.
        Author of the pipeline

        :return: The author of this ModelInput.
        :rtype: ArtifactAuthor
        """
        return self._author

    @author.setter
    def author(self, author):
        """
        Sets the author of this ModelInput.
        Author of the pipeline

        :param author: The author of this ModelInput.
        :type: ArtifactAuthor
        """

        self._author = author

    @property
    def pipeline_version_href(self):
        """
        Gets the pipeline_version_href of this ModelInput.
        Href of the pipeline version from which the model was created.

        :return: The pipeline_version_href of this ModelInput.
        :rtype: str
        """
        return self._pipeline_version_href

    @pipeline_version_href.setter
    def pipeline_version_href(self, pipeline_version_href):
        """
        Sets the pipeline_version_href of this ModelInput.
        Href of the pipeline version from which the model was created.

        :param pipeline_version_href: The pipeline_version_href of this ModelInput.
        :type: str
        """

        self._pipeline_version_href = pipeline_version_href

    @property
    def type(self):
        """
        Gets the type of this ModelInput.


        :return: The type of this ModelInput.
        :rtype: ModelType
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Sets the type of this ModelInput.


        :param type: The type of this ModelInput.
        :type: ModelType
        """

        self._type = type

    @property
    def runtime_environment(self):
        """
        Gets the runtime_environment of this ModelInput.


        :return: The runtime_environment of this ModelInput.
        :rtype: RuntimeEnvironment
        """
        return self._runtime_environment

    @runtime_environment.setter
    def runtime_environment(self, runtime_environment):
        """
        Sets the runtime_environment of this ModelInput.


        :param runtime_environment: The runtime_environment of this ModelInput.
        :type: RuntimeEnvironment
        """

        self._runtime_environment = runtime_environment

    @property
    def training_data_schema(self):
        """
        Gets the training_data_schema of this ModelInput.
        JSON schema of the training data

        :return: The training_data_schema of this ModelInput.
        :rtype: TrainingDataSchema
        """
        return self._training_data_schema

    @training_data_schema.setter
    def training_data_schema(self, training_data_schema):
        """
        Sets the training_data_schema of this ModelInput.
        JSON schema of the training data

        :param training_data_schema: The training_data_schema of this ModelInput.
        :type: TrainingDataSchema
        """

        self._training_data_schema = training_data_schema

    @property
    def label_col(self):
        """
        Gets the label_col of this ModelInput.
        Name of the label column

        :return: The label_col of this ModelInput.
        :rtype: str
        """
        return self._label_col

    @label_col.setter
    def label_col(self, label_col):
        """
        Sets the label_col of this ModelInput.
        Name of the label column

        :param label_col: The label_col of this ModelInput.
        :type: str
        """

        self._label_col = label_col

    @property
    def input_data_schema(self):
        """
        Gets the input_data_schema of this ModelInput.
        JSON schema of the input data (if not provided the training data schema is used)

        :return: The input_data_schema of this ModelInput.
        :rtype: InputDataSchema
        """
        return self._input_data_schema

    @input_data_schema.setter
    def input_data_schema(self, input_data_schema):
        """
        Sets the input_data_schema of this ModelInput.
        JSON schema of the input data (if not provided the training data schema is used)

        :param input_data_schema: The input_data_schema of this ModelInput.
        :type: InputDataSchema
        """

        self._input_data_schema = input_data_schema

    @property
    def output_data_schema(self):
        """
        Gets the output_data_schema of this ModelInput.
        JSON schema of the output data

        :return: The output_data_schema of this ModelInput.
        :rtype: OutputDataSchema
        """
        return self._output_data_schema

    @output_data_schema.setter
    def output_data_schema(self, output_data_schema):
        """
        Sets the output_data_schema of this ModelInput.
        JSON schema of the output data

        :param output_data_schema: The output_data_schema of this ModelInput.
        :type: OutputDataSchema
        """

        self._output_data_schema = output_data_schema

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.libraries_artifact_loader import LibrariesArtifactLoader


class LibrariesLoader(LibrariesArtifactLoader):
    """
        Returns  Libraries instance associated with this library artifact.
        :return: library zip file
    """
    def download_library(self, file_path):
        return self.load(file_path)





#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
"""
 An enumeration class for the crypto type field which describe what
 kind of data is being sent inside the Message
"""
from enum import Enum


class CryptoEnum(Enum):
    """
    Crypto types used for secure aggregation
    """

    CRYPTO_PAILLIER = "Paillier"
    CRYPTO_THRESHOLD_PAILLIER = "ThresholdPaillier"
    CRYPTO_MIFE = "MIFE"
    CRYPTO_MCFE = "MCFE"
    CRYPTO_DECENTRALIZED_MIFE = "DMIFE"
    CRYPTO_FHE = "FHE"

    KEY_PUBLIC_PARAMETER = "pp"
    KEY_PRIVATE = "sk"
    KEY_DECTYPT = "dk"
    WEIGHTS_PLAINTEXT = "plaintext-weights"
    WEIGHTS_CIPHERTEXT = "ciphertext-weights"



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import abc


class CryptoAsym(abc.ABC):
    """
    This class defines an interface for asymmetric encryption functions.
    """

    @abc.abstractmethod
    def __init__(self, key_file: str = None, password: bytes = None, **kwargs):
        return

    @abc.abstractstaticmethod
    def generate_key():
        raise NotImplementedError

    @abc.abstractmethod
    def get_public_key(self, type: str):
        raise NotImplementedError

    @abc.abstractmethod
    def write_key_file(self, file_path: str, password: bytes):
        raise NotImplementedError

    @abc.abstractmethod
    def encrypt(self, plain_data: bytes) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def decrypt(self, cipher_data: bytes) -> bytes:
        raise NotImplementedError

    @abc.abstractstaticmethod
    def encrypt_wkey(public_key, plain_data: bytes) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    def get_signature(self, data: bytes) -> bytes:
        raise NotImplementedError

    @abc.abstractstaticmethod
    def verify_signature(public_key, signature: bytes, data: bytes) -> bool:
        raise NotImplementedError



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from enum import Enum

__all__ = [
    "ModelTypes",
    "DecodingMethods",
]


class ModelTypes(Enum):
    """Supported foundation models."""
    FLAN_T5_XXL = "google/flan-t5-xxl"
    FLAN_UL2 = "google/flan-ul2"
    MT0_XXL = "bigscience/mt0-xxl"
    GPT_NEOX = 'eleutherai/gpt-neox-20b'
    MPT_7B_INSTRUCT2 = 'ibm/mpt-7b-instruct2'


class DecodingMethods(Enum):
    """Supported decoding methods for text generation."""
    SAMPLE = "sample"
    GREEDY = "greedy"



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import os
import shutil
import logging

from ibm_watson_machine_learning.libs.repo.mlrepository.artifact_reader import ArtifactReader
from ibm_watson_machine_learning.libs.repo.util.compression_util import CompressionUtil
from ibm_watson_machine_learning.libs.repo.util.unique_id_gen import uid_generate
from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *

lib_checker = LibraryChecker()

logger = logging.getLogger('TensorflowPipelineReader')


class TensorflowPipelineReader(ArtifactReader):
    def __init__(self, tensorflow_pipeline,
                 signature_def_map,
                 tags,
                 assets_collection,
                 legacy_init_op,
                 clear_devices,
                 main_op):
        self.archive_path = None
        self.tensorflow_pipeline = tensorflow_pipeline
        self.signature_def_map = signature_def_map
        self.tags = tags
        self.assets_collection = assets_collection
        self.legacy_init_op = legacy_init_op
        self.clear_devices = clear_devices
        self.main_op = main_op
        self.type_name = 'model'

    def read(self):
        return self._open_stream()

    def close(self):
        os.remove(self.archive_path)
        self.archive_path = None

    def _save_pipeline_archive(self):
        id_length = 20
        gen_id = uid_generate(id_length)
        temp_dir_name = '{}'.format(self.type_name + gen_id)
        temp_dir = os.path.join('.', temp_dir_name)
        self._save_tensorflow_model_to_dir(temp_dir)
        archive_path = self._compress_artifact(temp_dir, gen_id)
        shutil.rmtree(temp_dir)
        return archive_path

    def _compress_artifact(self, compress_artifact, gen_id):
        tar_filename = '{}_content.tar'.format(self.type_name + gen_id)
        gz_filename = '{}.gz'.format(tar_filename)
        CompressionUtil.create_tar(compress_artifact, '.', tar_filename)
        CompressionUtil.compress_file_gzip(tar_filename, gz_filename)
        os.remove(tar_filename)
        return gz_filename

    def _open_stream(self):
        if self.archive_path is None:
            self.archive_path = self._save_pipeline_archive()
        return open(self.archive_path, 'rb')

    def _save_tensorflow_model_to_dir(self, path):
        lib_checker.check_lib(TENSORFLOW)
        import tensorflow as tf
        try:
            from tensorflow import logging
            logging.set_verbosity(logging.WARN)
            if '2.1.0' in tf.__version__ :
                builder = tf.compat.v1.saved_model.builder.SavedModelBuilder(path)
            else:
                builder = tf.saved_model.builder.SavedModelBuilder(path)
            builder.add_meta_graph_and_variables(sess=self.tensorflow_pipeline,
                                                 tags=self.tags,
                                                 signature_def_map=self.signature_def_map,
                                                 assets_collection=self.assets_collection,
                                                 legacy_init_op=self.legacy_init_op,
                                                 clear_devices=self.clear_devices,
                                                 main_op=self.main_op)
            builder.save()
        except Exception as e:
            logMsg = "Tensorflow model Save failed with exception " + str(e)
            logger.info(logMsg)
            raise e




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2019- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import DATA_ASSETS_DETAILS_TYPE, modify_details_for_script_and_shiny
from ibm_watson_machine_learning.metanames import ScriptMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure, ForbiddenActionForGitBasedProject
import os
import warnings

_DEFAULT_LIST_LENGTH = 50


class Script(WMLResource):
    """Store and manage scripts assets."""

    ConfigurationMetaNames = ScriptMetaNames()
    """MetaNames for script assets creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    def get_details(self, script_uid=None):
        """Get script asset details. If no script_uid is passed, details for all script assets will be returned.

        :param script_uid: Unique id of script
        :type script_uid: str, optional

        :return: metadata of the stored script asset
        :rtype:
            - **dict** - if runtime_uid is not None
            - **{"resources": [dict]}** - if runtime_uid is None

        **Example**

        .. code-block:: python

            script_details = client.script.get_details(script_uid)

        """
        def get_required_elements(response):
            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                return response
            else:
                response = modify_details_for_script_and_shiny(response)
                final_response = {
                    "metadata": response[u'metadata'],
                }

                if 'entity' in response:
                    final_response['entity'] = response[u'entity']

                    try:
                        del final_response['entity'][u'script'][u'ml_version']
                    except KeyError:
                        pass

                return final_response

        return self._get_asset_based_resource(script_uid, 'script', get_required_elements)

    def store(self, meta_props, file_path):
        """Create a script asset and upload content to it.

        :param meta_props: name to be given to the script asset
        :type meta_props: str
        :param file_path: path to the content file to be uploaded
        :type file_path: str

        :return: metadata of the stored script asset
        :rtype: dict

        **Example**

        .. code-block:: python

            metadata = {
                client.script.ConfigurationMetaNames.NAME: 'my first script',
                client.script.ConfigurationMetaNames.DESCRIPTION: 'description of the script',
                client.script.ConfigurationMetaNames.SOFTWARE_SPEC_UID: '0cdb0f1e-5376-4f4d-92dd-da3b69aa9bda'
            }

            asset_details = client.script.store(meta_props=metadata, file_path="/path/to/file")

        """
        if self._client.project_type == 'local_git_storage':
            raise ForbiddenActionForGitBasedProject(reason="Storing Scripts is not supported for git based project.")


        WMLResource._chk_and_block_create_update_for_python36(self)
        #Script._validate_type(name, u'name', str, True)
        Script._validate_type(file_path, u'file_path', str, True)
        script_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client
        )

        name, extension = os.path.splitext(file_path)

        response = self._create_asset(script_meta, file_path, extension)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:

            entity = response[u'entity']

            try:
                del entity[u'script'][u'ml_version']
            except KeyError:
                pass

            final_response = {
                "metadata": response[u'metadata'],
                "entity": entity
            }

            return final_response

    def _create_asset(self, script_meta, file_path, extension='.py'):

        ##Step1: Create a data asset
        name = script_meta['metadata'][u'name']

        if script_meta['metadata'].get('description') is not None:
            desc = script_meta['metadata'][u'description']
        else:
            desc = ""

        if script_meta.get('software_spec_uid') is not None:
            if script_meta[u'software_spec_uid'] == "":
                raise WMLClientError("Failed while creating a script asset, SOFTWARE_SPEC_UID cannot be empty")
                return

        base_script_asset = {
            "description": "Script File",
            "fields": [
                {
                    "key": "language",
                    "type": "string",
                    "facet": False,
                    "is_array": False,
                    "search_path": "asset.type",
                    "is_searchable_across_types": True
                }
            ],
            "relationships": [
                {
                    "key": "software_spec.id",
                    "target_asset_type": "software_specification",
                    "on_delete_target": "IGNORE",
                    "on_delete": "IGNORE",
                    "on_clone_target": "IGNORE"
                }
            ],
            "name": "script",
            "version": 1
        }

        if extension == '.py':
            lang = 'python3'
        elif extension == '.R':
            lang = 'R'
        else:
            raise WMLClientError("This file type is not supported. It has to be either a python script(.py file ) or a "
                                 "R script")

        #check if the software spec specified is base or derived and
        # accordingly update the entity for asset creation
        # TODO WSD2.0
        #kaganesa: remove the below if and else. once v2/software_spec apis are available in WSD2.0,
        # everything inside if should work.
        if not self._client.WSD:
            sw_spec_details = self._client.software_specifications.get_details(script_meta[u'software_spec_uid'])

            if lang == 'R':
                if self._client.ICP_47:
                    r_sw_specs = []
                    for sw_spec_details in self._client.software_specifications.get_details()['resources']:
                        sw_configuration = sw_spec_details['entity'].get('software_specification', {}).get(
                            'software_configuration', {})
                        if sw_configuration.get('platform', {}).get('name') == 'r' and len(
                                sw_configuration.get('included_packages', [])) > 1:
                            r_sw_specs.append(sw_spec_details['metadata']['name'])
                    rscript_sw_specs = []               # names of supported r script software specifications
                    deprecated_rscript_sw_specs = []    # names of deprecated r script software specifications
                    for sw_spec in self._client.software_specifications.get_details(state_info=True)['resources']:
                        if sw_spec['name'] in r_sw_specs:
                            if sw_spec['state'] == 'supported':
                                rscript_sw_specs.append(sw_spec['name'])
                            elif sw_spec['state'] == 'deprecated':
                                deprecated_rscript_sw_specs.append(sw_spec['name'])

                elif self._client.ICP_46:
                    rscript_sw_specs = ('runtime-22.2-r4.2',)
                    deprecated_rscript_sw_specs = ('default_r3.6', 'runtime-22.1-r3.6',)
                else:
                    rscript_sw_specs = ('default_r3.6', 'runtime-22.1-r3.6',)
                    deprecated_rscript_sw_specs = ()
                rscript_sw_spec_ids = [self._client.software_specifications.get_id_by_name(sw_name)
                                       for sw_name in rscript_sw_specs]
                deprecated_sw_spec_ids = [self._client.software_specifications.get_id_by_name(sw_name)
                                       for sw_name in deprecated_rscript_sw_specs]

                if script_meta[u'software_spec_uid'] not in rscript_sw_spec_ids + deprecated_sw_spec_ids:
                    raise WMLClientError(f"For R scripts, only base software specs {','.join(rscript_sw_specs)} "
                                         "are supported. Specify "
                                         "the id you get via "
                                         "self._client.software_specifications.get_id_by_name(sw_name)")
                elif script_meta[u'software_spec_uid'] in deprecated_rscript_sw_specs:
                    warnings.warn("Provided software spec is deprecated for R scripts. "
                             f"Only base software specs {','.join(rscript_sw_specs)} "
                             "are supported. Specify "
                             "the id you get via "
                             "self._client.software_specifications.get_id_by_name(sw_name)")


            if(sw_spec_details[u'entity'][u'software_specification'][u'type'] == 'base'):
                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                    asset_meta = {
                        "metadata": {
                            "name": name,
                            "description": desc,
                            "asset_type": "script",
                            "origin_country": "us",
                            "asset_category": "USER"
                        },
                        "entity": {
                            "script": {
                                "language": {
                                    # "name": "python3"
                                    "name": lang
                    },
                                "software_spec": {
                                    "base_id": script_meta[u'software_spec_uid']
                                }
                             }
                            }
                    }
                else:
                     asset_meta = {
                         "metadata": {
                             "name": name,
                             "description": desc,
                             "asset_type": "script",
                             "origin_country": "us",
                             "asset_category": "USER"
                         },
                         "entity": {
                             "script": {
                                 "ml_version": "4.0.0",
                                 "language": {
                                     # "name": "python3"
                                     "name": lang
                     },
                                 "software_spec": {
                                     "base_id": script_meta[u'software_spec_uid']
                                 }
                             }
                         }
                     }
            else:
                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                    asset_meta = {
                        "metadata": {
                            "name": name,
                            "description": desc,
                            "asset_type": "script",
                            "origin_country": "us",
                            "asset_category": "USER"
                        },
                        "entity": {
                            "script": {
                                "language": {
                                    # "name": "python3"
                                    "name": lang
                    },
                                "software_spec": {
                                    "id": script_meta[u'software_spec_uid']
                                }
                            }
                        }
                    }
                else:
                    asset_meta = {
                        "metadata": {
                            "name": name,
                            "description": desc,
                            "asset_type": "script",
                            "origin_country": "us",
                            "asset_category": "USER"
                        },
                        "entity": {
                            "script": {
                                "ml_version": "4.0.0",
                                "language": {
                                    # "name": "python3"
                                    "name": lang
                                },
                                "software_spec": {
                                    "id": script_meta[u'software_spec_uid']
                                }
                            }
                        }
                    }
        else:
            asset_meta = {
                "metadata": {
                    "name": name,
                    "description": desc,
                    "asset_type": "script",
                    "origin_country": "us",
                    "asset_category": "USER"
                },
                "entity": {
                    "script": {
                        "language": {
                            # "name": "python3"
                            "name": lang
            },
                        "software_spec": {
                            "id": script_meta[u'software_spec_uid']
                        }
                    }
                }
            }

        #Step1  : Create an asset
        print("Creating Script asset...")

        if self._client.WSD:
            # For WSD the asset creation is done within _wsd_create_asset function using polyglot
            # Thus using the same for data_assets type


            meta_props = {
                    "name": name
            }
            details = Script._wsd_create_asset(self, "script", asset_meta, meta_props, file_path, user_archive_file=True)
            return self._get_required_element_from_response(details)
        else:
            if not self._ICP:
                creation_response = requests.post(
                        self._client.service_instance._href_definitions.get_assets_href(),
                        headers=self._client._get_headers(),
                        params = self._client._params(),
                        json=asset_meta
                )
            else:
                if self._client.ICP_PLATFORM_SPACES:
                    creation_response = requests.post(
                        self._client.service_instance._href_definitions.get_data_assets_href(),
                        headers=self._client._get_headers(),
                        json=asset_meta,
                        params=self._client._params()
                    )
                else:
                    asset_type_response = requests.post(
                        self._client.service_instance._href_definitions.get_wsd_asset_type_href() + '?',
                        headers=self._client._get_headers(),
                        json=base_script_asset,
                        params=self._client._params()
                    )
                    if asset_type_response.status_code == 201 or asset_type_response.status_code == 409:
                        creation_response = requests.post(
                            self._client.service_instance._href_definitions.get_data_assets_href(),
                            headers=self._client._get_headers(),
                            json=asset_meta,
                            params=self._client._params()
                        )

            asset_details = self._handle_response(201, u'creating new asset', creation_response)
            #Step2: Create attachment
            if creation_response.status_code == 201:
                asset_id = asset_details["metadata"]["asset_id"]
                attachment_meta = {
                        "asset_type": "script",
                        "name": "attachment_"+asset_id
                    }

                attachment_response = requests.post(
                    self._client.service_instance._href_definitions.get_attachments_href(asset_id),
                    headers=self._client._get_headers(),
                    params=self._client._params(),
                    json=attachment_meta
                )
                attachment_details = self._handle_response(201, u'creating new attachment', attachment_response)
                if attachment_response.status_code == 201:
                    attachment_id = attachment_details["attachment_id"]
                    attachment_url = attachment_details["url1"]

                    #Step3: Put content to attachment
                    try:
                        with open(file_path, 'rb') as f:
                            if not self._ICP:
                                put_response = requests.put(
                                    attachment_url,
                                    data=f.read()
                                )
                            else:
                                put_response = requests.put(
                                    self._wml_credentials['url'] + attachment_url,
                                    files={'file': (name, f, 'application/octet-stream')}
                                )
                    except Exception as e:
                        deletion_response = requests.delete(
                            self._client.service_instance._href_definitions.get_data_asset_href(asset_id),
                            params=self._client._params(),
                            headers=self._client._get_headers()
                        )
                        print(deletion_response.status_code)
                        raise WMLClientError("Failed while reading a file.", e)

                    if put_response.status_code == 201 or put_response.status_code == 200:
                        # Step4: Complete attachment
                        complete_response = requests.post(
                            self._client.service_instance._href_definitions.get_attachment_complete_href(asset_id, attachment_id),
                            headers=self._client._get_headers(),
                            params = self._client._params()

                        )

                        if complete_response.status_code == 200:
                            print("SUCCESS")
                            return self._get_required_element_from_response(asset_details)
                        else:
                            self._delete(asset_id)
                            raise WMLClientError("Failed while creating a script asset. Try again.")
                    else:
                        self._delete(asset_id)
                        raise WMLClientError("Failed while creating a script asset. Try again.")
                else:
                    print("SUCCESS")
                    return self._get_required_element_from_response(asset_details)
            else:
                raise WMLClientError("Failed while creating a script asset. Try again.")


    def list(self, limit=None, return_as_df=True):
        """Print stored scripts in a table format. If limit is set to None there will be only first 50 records shown.

        :param limit: limit number of fetched records
        :type limit: int, optional
        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed scripts or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        **Example**

        .. code-block:: python

            client.script.list()
        """

        Script._validate_type(limit, u'limit', int, False)
        href = self._client.service_instance._href_definitions.get_search_script_href()

        data = {
                "query": "*:*"
        }
        if limit is not None:
            data.update({"limit": limit})

        response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(),json=data)

        self._handle_response(200, u'list assets', response)
        asset_details = self._handle_response(200, u'list assets', response)["results"]
        space_values = [
            (m[u'metadata'][u'name'], m[u'metadata'][u'asset_type'], m["metadata"]["asset_id"]) for
            m in asset_details]

        table = self._list(space_values, [u'NAME', u'ASSET_TYPE', u'ASSET_ID'], None, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table

    def download(self, asset_uid, filename, rev_uid=None):
        """Download the content of a script asset.

        :param asset_uid: the Unique Id of the script asset to be downloaded
        :type asset_uid: str
        :param filename: filename to be used for the downloaded file
        :type filename: str
        :param rev_uid: revision id
        :type rev_uid: str, optional

        :return: path to the downloaded asset content
        :rtype: str

        **Example**

        .. code-block:: python

            client.script.download(asset_uid, "script_file")
        """
        if rev_uid is not None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not applicable for this release')

        Script._validate_type(asset_uid, u'asset_uid', str, True)
        Script._validate_type(rev_uid, u'rev_uid', int, False)

        params = self._client._params()

        if rev_uid is not None:
            params.update({'revision_id': rev_uid})

        import urllib
        if not self._ICP:
            asset_response = requests.get(self._client.service_instance._href_definitions.get_asset_href(asset_uid),
                                          params=params,
                                          headers=self._client._get_headers())
        else:
            asset_response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(asset_uid),
                                          params=params,
                                          headers=self._client._get_headers())
        asset_details = self._handle_response(200, u'get assets', asset_response)

        if self._WSD:
            attachment_url = asset_details['attachments'][0]['object_key']
            artifact_content_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                   urllib.parse.quote('script/' + attachment_url, safe='')

            r = requests.get(artifact_content_url, params=self._client._params(), headers=self._client._get_headers(),
                             stream=True)
            if r.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading data asset"), r)

            downloaded_asset = r.content
            try:
                with open(filename, 'wb') as f:
                    f.write(downloaded_asset)
                print(u'Successfully saved data asset content to file: \'{}\''.format(filename))
                return os.path.abspath(filename)
            except IOError as e:
                raise WMLClientError(u'Saving data asset with artifact_url: \'{}\'  to local file failed.'.format(filename), e)
        else:
            attachment_id = asset_details["attachments"][0]["id"]

            response = requests.get(self._client.service_instance._href_definitions.get_attachment_href(asset_uid,attachment_id), params=params,
                                    headers=self._client._get_headers())

            if response.status_code == 200:
                attachment_signed_url = response.json()["url"]
                if 'connection_id' in asset_details["attachments"][0]:
                    att_response = requests.get(attachment_signed_url)
                else:
                    if not self._ICP:
                        att_response = requests.get(attachment_signed_url)
                    else:
                        att_response = requests.get(self._wml_credentials["url"]+attachment_signed_url)
                if att_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading asset"), att_response)

                downloaded_asset = att_response.content
                try:
                    with open(filename, 'wb') as f:
                        f.write(downloaded_asset)
                    print(u'Successfully saved data asset content to file: \'{}\''.format(filename))
                    return os.path.abspath(filename)
                except IOError as e:
                    raise WMLClientError(u'Saving asset with artifact_url to local file: \'{}\' failed.'.format(filename), e)
            else:
                raise WMLClientError("Failed while downloading the asset " + asset_uid)

    @staticmethod
    def get_id(asset_details):
        """Get Unique Id of stored script asset.

        :param asset_details: metadata of the stored script asset
        :type asset_details: dict

        :return: Unique Id of stored script asset
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.script.get_id(asset_details)
        """

        return Script.get_uid(asset_details)

    @staticmethod
    def get_uid(asset_details):
        """Get Unique Id  of stored script asset.

        *Deprecated:* Use ``get_id(asset_details)`` instead.

        :param asset_details: metadata of the stored script asset
        :type asset_details: dict

        :return: Unique Id of stored script asset
        :rtype: str

        **Example**

        .. code-block:: python

            asset_uid = client.script.get_uid(asset_details)
        """
        Script._validate_type(asset_details, u'asset_details', object, True)
        Script._validate_type_of_details(asset_details, DATA_ASSETS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(asset_details, u'data_assets_details',
                                                           [u'metadata', u'guid'])

    @staticmethod
    def get_href(asset_details):
        """Get url of stored scripts asset.

        :param asset_details: stored script details
        :type asset_details: dict

        :return: href of stored script asset
        :rtype: str

        **Example**

        .. code-block:: python

            asset_details = client.script.get_details(asset_uid)
            asset_href = client.script.get_href(asset_details)

        """
        Script._validate_type(asset_details, u'asset_details', object, True)
        Script._validate_type_of_details(asset_details, DATA_ASSETS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(asset_details, u'asset_details', [u'metadata', u'href'])

    def update(self, script_uid, meta_props=None, file_path=None):
        """Update script with either metadata or attachment or both.

        :param script_uid: script UID
        :type script_uid: str
        :param meta_props: changes for script matadata
        :type meta_props: dict, optional
        :param file_path: file path to new attachment
        :type file_path: str, optional

        :return: updated metadata of script
        :rtype: dict

        **Example**

        .. code-block:: python

            script_details = client.script.update(model_uid, meta, content_path)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        # We need to enable this once we add functionality for WSD
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        Script._validate_type(script_uid, 'script_uid', str, True)

        if meta_props is None and file_path is None:
            raise WMLClientError('Atleast either meta_props or file_path has to be provided')

        updated_details = None
        details = None

        url = self._client.service_instance._href_definitions.get_asset_href(script_uid)

        # STEPS
        # STEP 1. Get existing metadata
        # STEP 2. If meta_props provided, we need to patch meta
        #   CAMS has meta and entity patching. 'name' and 'description' get stored in CAMS meta section
        #   a. Construct meta patch string and call /v2/assets/<asset_id> to patch meta
        #   b. Construct entity patch if required and call /v2/assets/<asset_id>/attributes/script to patch entity
        # STEP 3. If file_path provided, we need to patch the attachment
        #   a. If attachment already exists for the script, delete it
        #   b. POST call to get signed url for upload
        #   c. Upload to the signed url
        #   d. Mark upload complete
        # STEP 4. Get the updated script record and return

        # STEP 1
        response = requests.get(
            url,
            params=self._client._params(),
            headers=self._client._get_headers()
        )

        if response.status_code != 200:
            if response.status_code == 404:
                raise WMLClientError(
                    u'Invalid input. Unable to get the details of script_uid provided.')
            else:
                raise ApiRequestFailure(u'Failure during {}.'.format("getting script to update"), response)

        details = self._handle_response(200, "Get script details", response)

        attachments_response = None

        # STEP 2a.
        # Patch meta if provided
        if meta_props is not None:
            self._validate_type(meta_props, u'meta_props', dict, True)

            meta_patch_payload = []
            entity_patch_payload = []

            # Since we are dealing with direct asset apis, there can be metadata or entity patch or both
            if "name" in meta_props or "description" in meta_props:
                props_for_asset_meta_patch = {}

                for key in meta_props:
                    if key == 'name' or key == 'description':
                        props_for_asset_meta_patch.update({key: meta_props[key]})

                meta_patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details,
                                                                                         props_for_asset_meta_patch,
                                                                                         with_validation=True)
            # STEP 2b.
            if "software_spec_uid" in meta_props:
                if details[u'entity'][u'script'][u'software_spec']:
                    entity_patch_payload = [{'op':'replace',
                                             'path': '/software_spec/base_id',
                                             'value': meta_props[u'software_spec_uid']}]
                else:
                    entity_patch_payload = [{'op': 'add',
                                             'path': '/software_spec',
                                             'value': '{base_id:' + meta_props[u'software_spec_uid'] + '}'}]

            if meta_patch_payload:
                meta_patch_url = self._client.service_instance._href_definitions.get_asset_href(script_uid)

                response_patch = requests.patch(meta_patch_url,
                                                json=meta_patch_payload,
                                                params=self._client._params(),
                                                headers=self._client._get_headers())

                updated_details = self._handle_response(200, u'script patch', response_patch)

            if entity_patch_payload:
                entity_patch_url = self._client.service_instance._href_definitions.get_asset_href(script_uid) + '/attributes/script'

                response_patch = requests.patch(entity_patch_url,
                                                json=entity_patch_payload,
                                                params=self._client._params(),
                                                headers=self._client._get_headers())

                updated_details = self._handle_response(200, u'script patch', response_patch)

        if file_path is not None:
            if "attachments" in details and details[u'attachments']:
                current_attachment_id = details[u'attachments'][0][u'id']
            else:
                current_attachment_id = None

            #STEP 3
            attachments_response = self._update_attachment_for_assets("script",
                                                                      script_uid,
                                                                      file_path,
                                                                      current_attachment_id)

        if attachments_response is not None and 'success' not in attachments_response:
            self._update_msg(updated_details)

        # Have to fetch again to reflect updated asset and attachment ids
        url = self._client.service_instance._href_definitions.get_asset_href(script_uid)

        response = requests.get(
            url,
            params=self._client._params(),
            headers=self._client._get_headers()
        )

        if response.status_code != 200:
            if response.status_code == 404:
                raise WMLClientError(
                    u'Invalid input. Unable to get the details of script_uid provided.')
            else:
                raise ApiRequestFailure(u'Failure during {}.'.format("getting script to update"), response)

        # response = self._handle_response(200, "Get script details", response)

        # return self._get_required_element_from_response(response)

        response = self._get_required_element_from_response(self._handle_response(200, "Get script details", response))

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            entity = response[u'entity']

            try:
                del entity[u'script'][u'ml_version']
            except KeyError:
                pass

            final_response = {
                "metadata": response[u'metadata'],
                "entity": entity
            }

            return final_response

    def _update_msg(self, updated_details):
        if updated_details is not None:
            print("Could not update the attachment because of server error."
                  " However metadata is updated. Try updating attachment again later")
        else:
            raise WMLClientError('Unable to update attachment because of server error. Try again later')

    def delete(self, asset_uid):
        """Delete a stored script asset.

        :param asset_uid: UID of script asset
        :type asset_uid: str

        :return: status ("SUCCESS" or "FAILED")
        :rtype: str

        **Example**

        .. code-block:: python

            client.script.delete(asset_uid)

        """
        Script._validate_type(asset_uid, u'asset_uid', str, True)
        if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and \
                self._if_deployment_exist_for_asset(asset_uid):
            raise WMLClientError(
                u'Cannot delete script that has existing deployments. Please delete all associated deployments and try again')

        response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(asset_uid), params=self._client._params(),
                                headers=self._client._get_headers())

        if response.status_code == 200:
            return self._get_required_element_from_response(response.json())
        else:
            return self._handle_response(204, u'delete assets', response)

    def create_revision(self, script_uid):
        """Create revision for the given script. Revisions are immutable once created.
        The metadata and attachment at `script_uid` is taken and a revision is created out of it.

        :param script_uid: script ID
        :type script_uid: str

        :return: stored script revisions metadata
        :rtype: dict

        **Example**

        .. code-block:: python

            script_revision = client.script.create_revision(script_uid)
        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        # We need to enable this once we add functionality for WSD
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()
        Script._validate_type(script_uid, u'script_uid', str, True)

        print("Creating script revision...")

        # return self._get_required_element_from_response(
        #     self._create_revision_artifact_for_assets(script_uid, 'Script'))

        response = self._get_required_element_from_response(
            self._create_revision_artifact_for_assets(script_uid, 'Script'))

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            entity = response[u'entity']

            try:
                del entity[u'script'][u'ml_version']
            except KeyError:
                pass

            final_response = {
                "metadata": response[u'metadata'],
                "entity": entity
            }

            return final_response

    def list_revisions(self, script_uid, limit=None, return_as_df=True):
        """Print all revisions for the given script uid in a table format.

        :param script_uid: stored script ID
        :type script_uid: str
        :param limit: limit number of fetched records
        :type limit: int, optional
        :param return_as_df: determinate if table should be returned as pandas.DataFrame object, default: True
        :type return_as_df: bool, optional

        :return: pandas.DataFrame with listed revisions or None if return_as_df is False
        :rtype: pandas.DataFrame or None

        **Example**

        .. code-block:: python

            client.script.list_revisions(script_uid)
        """
        # We need to enable this once we add functionality for WSD
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()

        Script._validate_type(script_uid, u'script_uid', str, True)

        url = self._client.service_instance._href_definitions.get_asset_href(script_uid) + "/revisions"
        # /v2/assets/{asset_id}/revisions returns 'results' object
        script_resources = self._get_with_or_without_limit(url,
                                                           limit,
                                                           'List Script revisions',
                                                           summary=None,
                                                           pre_defined=None)[u'resources']
        script_values = [
            (m[u'metadata'][u'asset_id'],
             m[u'metadata'][u'revision_id'],
             m[u'metadata'][u'name'],
             m[u'metadata'][u'commit_info'][u'committed_at']) for m in
            script_resources]

        table = self._list(script_values, [u'GUID', u'REVISION_ID', u'NAME', u'REVISION_COMMIT'], limit, _DEFAULT_LIST_LENGTH)
        if return_as_df:
            return table

    def get_revision_details(self, script_uid, rev_uid=None):
        """Get metadata of script revision.

        :param script_uid: script ID
        :type script_uid: str
        :param rev_uid: revision ID, if this parameter is not provided, returns latest revision if existing else error
        :type rev_uid: int, optional

        :return: stored script(s) metadata
        :rtype: dict

        **Example**

        .. code-block:: python

            script_details = client.script.get_revision_details(script_uid, rev_uid)
        """
        # We need to enable this once we add functionality for WSD
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()

        Script._validate_type(script_uid, u'script_uid', str, True)
        Script._validate_type(rev_uid, u'rev_uid', int, False)

        if rev_uid is None:
            rev_uid = 'latest'

        url = self._client.service_instance._href_definitions.get_asset_href(script_uid)
        # return self._get_required_element_from_response(self._get_with_or_without_limit(url,
        #                                        limit=None,
        #                                        op_name="asset_revision",
        #                                        summary=None,
        #                                        pre_defined=None,
        #                                        revision=rev_uid))
        resources = self._get_with_or_without_limit(url,
                                                    limit=None,
                                                    op_name="asset_revision",
                                                    summary=None,
                                                    pre_defined=None,
                                                    revision=rev_uid)['resources']
        responses = [self._get_required_element_from_response(resource) for resource in resources]

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return responses
        else:
            final_responses = []
            for response in responses:
                entity = response[u'entity']

                try:
                    del entity[u'script'][u'ml_version']
                except KeyError:
                    pass

                final_responses.append({
                    "metadata": response[u'metadata'],
                    "entity": entity
                })

            return final_responses

    def _delete(self, asset_uid):
        Script._validate_type(asset_uid, u'asset_uid', str, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(asset_uid), params=self._client._params(),
                                   headers=self._client._get_headers())

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'scripts', dict)

        revision_id = None

        try:
            if self._client.default_space_id is not None:
                metadata = {'space_id': response_data['metadata']['space_id'],
                            'name':response_data['metadata']['name'],
                            'guid': response_data['metadata']['asset_id'],
                            'href': response_data['href'],
                            'asset_type': response_data['metadata']['asset_type'],
                            'created_at': response_data['metadata']['created_at'],
                            'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                            }
                if 'description' in response_data[u'metadata']:
                    metadata.update(
                        {'description':response_data[u'metadata'][u'description']})

                if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    if "revision_id" in response_data[u'metadata']:
                        revision_id = response_data[u'metadata'][u'revision_id']
                        metadata.update({'revision_id': response_data[u'metadata'][u'revision_id']})

                    if "attachments" in response_data and response_data[u'attachments']:
                        metadata.update({'attachment_id': response_data[u'attachments'][0][u'id']})

                    if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                        metadata.update(
                            {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                new_el = {'metadata': metadata,
                          'entity': response_data['entity']
                          }
            elif self._client.default_project_id is not None:
                if self._client.WSD:

                    href = self._client.service_instance._href_definitions.get_base_asset_href(response_data['metadata']['asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']

                    metadata = {'project_id': response_data['metadata']['project_id'],
                                'guid': response_data['metadata']['asset_id'],
                                'name': response_data['metadata']['name'],
                                'href': href,
                                'asset_type': response_data['metadata']['asset_type'],
                                'created_at': response_data['metadata']['created_at']
                                }

                    if 'description' in response_data[u'metadata']:
                        metadata.update(
                            {'description': response_data[u'metadata'][u'description']})

                    new_el = {'metadata': metadata,
                              'entity': response_data['entity']
                              }
                    if self._client.WSD_20 is not None:
                        if "revision_id" in response_data[u'metadata']:
                            revision_id = response_data[u'metadata'][u'revision_id']
                            new_el['metadata'].update({'revision_id': response_data[u'metadata'][u'revision_id']})

                        if "attachments" in response_data and response_data[u'attachments']:
                            new_el['metadata'].update({'attachment_id': response_data[u'attachments'][0][u'id']})

                        if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                            new_el['metadata'].update(
                                {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                    if 'usage' in response_data['metadata']:
                        new_el['metadata'].update({'last_updated_at': response_data['metadata']['usage']['last_updated_at']})
                    else:
                        new_el['metadata'].update(
                         {'last_updated_at': response_data['metadata']['last_updated_at']})
                else:
                    metadata = {'project_id': response_data['metadata']['project_id'],
                                'guid': response_data['metadata']['asset_id'],
                                'href': response_data['href'],
                                'name': response_data['metadata']['name'],
                                'asset_type': response_data['metadata']['asset_type'],
                                'created_at': response_data['metadata']['created_at'],
                                'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                                }
                    if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                        if "revision_id" in response_data[u'metadata']:
                            revision_id = response_data[u'metadata'][u'revision_id']
                            metadata.update({'revision_id': response_data[u'metadata'][u'revision_id']})

                        if "attachments" in response_data and response_data[u'attachments']:
                            metadata.update({'attachment_id': response_data[u'attachments'][0][u'id']})

                        if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                            metadata.update(
                                {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                    if 'description' in response_data[u'metadata']:
                        metadata.update(
                            {'description': response_data[u'metadata'][u'description']})
                    new_el = {'metadata': metadata,
                              'entity': response_data['entity']
                              }
                    if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                        href_without_host = response_data['href'].split('.com')[-1]
                        new_el[u'metadata'].update({'href': href_without_host})
            return new_el
        except Exception as e:
            raise WMLClientError(f"Failed to read Response from down-stream service: {response_data}")



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class HyperParametersExperimentsDoubleRange(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, min_value=None, max_value=None, step=None, power=None):
        """
        HyperParametersExperimentsDoubleRange - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'min_value': 'float',
            'max_value': 'float',
            'step': 'float',
            'power': 'float'
        }

        self.attribute_map = {
            'min_value': 'min_value',
            'max_value': 'max_value',
            'step': 'step',
            'power': 'power'
        }

        self._min_value = min_value
        self._max_value = max_value
        self._step = step
        self._power = power

    @property
    def min_value(self):
        """
        Gets the min_value of this HyperParametersExperimentsDoubleRange.


        :return: The min_value of this HyperParametersExperimentsDoubleRange.
        :rtype: float
        """
        return self._min_value

    @min_value.setter
    def min_value(self, min_value):
        """
        Sets the min_value of this HyperParametersExperimentsDoubleRange.


        :param min_value: The min_value of this HyperParametersExperimentsDoubleRange.
        :type: float
        """

        self._min_value = min_value

    @property
    def max_value(self):
        """
        Gets the max_value of this HyperParametersExperimentsDoubleRange.


        :return: The max_value of this HyperParametersExperimentsDoubleRange.
        :rtype: float
        """
        return self._max_value

    @max_value.setter
    def max_value(self, max_value):
        """
        Sets the max_value of this HyperParametersExperimentsDoubleRange.


        :param max_value: The max_value of this HyperParametersExperimentsDoubleRange.
        :type: float
        """

        self._max_value = max_value

    @property
    def step(self):
        """
        Gets the step of this HyperParametersExperimentsDoubleRange.


        :return: The step of this HyperParametersExperimentsDoubleRange.
        :rtype: float
        """
        return self._step

    @step.setter
    def step(self, step):
        """
        Sets the step of this HyperParametersExperimentsDoubleRange.


        :param step: The step of this HyperParametersExperimentsDoubleRange.
        :type: float
        """

        self._step = step

    @property
    def power(self):
        """
        Gets the power of this HyperParametersExperimentsDoubleRange.


        :return: The power of this HyperParametersExperimentsDoubleRange.
        :rtype: float
        """
        return self._power

    @power.setter
    def power(self, power):
        """
        Sets the power of this HyperParametersExperimentsDoubleRange.


        :param power: The power of this HyperParametersExperimentsDoubleRange.
        :type: float
        """

        self._power = power

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



"""
Module providing utility functions helpful for both party and aggregator
"""

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import os
import yaml
import logging
import logging.config
import importlib
from importlib.machinery import SourceFileLoader
from pathlib import PurePath, Path

import ibmfl.envs as fl_envs
import ibmfl.util.log_config as log_config
from ibmfl.exceptions import InvalidConfigurationException, FLException


"""
Module providing utility functions helpful for both party and aggregator
"""
logger = logging.getLogger(__name__)


def read_yaml_config(config_file):
    """
    reads the yaml config file and generates a dictionary

    :param config_file: yaml file containing the definitions of formatter and
    handler
    :return : dictionary object containing the config details
    """
    if config_file and os.path.exists(config_file):
        with open(config_file, 'rt') as cf:
            try:
                return yaml.safe_load(cf.read())
            except InvalidConfigurationException as ice:
                logger.error('Failed to load yaml configurations. ' + str(ice))
                logging.error('Could not load yaml configuration')
    else:
        logger.info('Yaml configuration file not found: ' + str(config_file))

    return None


def configure_logging_from_file(config_file='log_config.yaml',
                                log_level='INFO'):
    """
    configures logging for application based on the configuration file

    :param config_file: yaml file containing the definitions of formatter and handler
    :param log_level: should be a value from [DEBUG, INFO, WARNING, ERROR, CRITICAL]
            based on the required granularity
    :return: a boolean object. False for default basic config True otherwise
    """
    log_level = logging.getLevelName(log_level)
    config = read_yaml_config(config_file)

    result = log_config.configure_logging(config, log_level)

    return result


def get_aggregator_config(**kwargs):
    """Retrieve connections settings from config file or arguments

    :param \\*\\*kwargs: Arguments that are passed into the aggregator instance
    """

    if not kwargs:
        raise InvalidConfigurationException(
            'No arguments given to Aggregator at runtime')

    config_file = kwargs.get('config_file')

    if config_file:
        return get_config_from_file(config_file=config_file)
    else:
        return get_config_from_args(**kwargs)


def get_party_config(**kwargs):
    """
    Retrieve connections settings from config file or arguments

    :param \\*\\*kwargs: Arguments that are passed into the party instance
    """

    if not kwargs:
        raise InvalidConfigurationException(
            'No arguments given to Party at runtime')

    config_file = kwargs.get('config_file')

    if config_file:
        logger.info('Getting config from file')
        return get_config_from_file(config_file=config_file)
    else:
        return get_config_from_args(**kwargs)


def get_authority_config(**kwargs):
    """
    Retrieve connections settings from config file or arguments

    :param \\*\\*kwargs: Arguments that are passed into the party instance
    """

    if not kwargs:
        raise InvalidConfigurationException(
            'No arguments given to Authority at runtime')

    config_file = kwargs.get('config_file')

    if config_file:
        logger.info('Getting config from file')
        return get_config_from_file(config_file=config_file)
    else:
        return get_config_from_args(**kwargs)


def get_config_from_file(config_file):
    """
    Reads a yaml file and resolves the string configuration to appropriate
    class instances.
    :param config_file: yaml file containing the configurations
    :type config_file: `str`
    :return: dictionary containing class references and additional information
    :rtype: `dict`
    """
    logger.info('Getting details from config file.')
    logger.debug('Reading Configuration File : %s', config_file)
    config_dict = read_yaml_config(config_file)

    cls_config = get_cls_by_config(config_dict)

    return cls_config


def get_config_from_args(config_dict=None, **kwargs):
    """
    Reads arguments and resolves the string configuration to appropriate
    class instances.
    :param config_file: yaml file containing the configurations
    :type config_file: `str`
    :return: dictionary of class references
    :rtype: `dict`
    """
    logger.info('Getting Aggregator details from arguments.')

    cls_config = {}
    if config_dict:
        cls_config = get_cls_by_config(config_dict)
    else:
        cls_config = get_cls_by_config(kwargs)

        # TODO: iterate through **kwargs and resolve any configurations given
        # as arguments to corresponding class
    return cls_config


def get_cls_by_config(config_dict):
    """
    Resolve the class name string references given in config file
    to its actual class reference
    :param config_dict: dictionary of configuration
    :type config_dict: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    cls_config = {}
    # TODO: since all the configuration blocks carry same signature this
    # should be done in iterative way for all keys in the dictionary
    cls_config['data'] = get_data_from_config(config_dict.get('data'))
    cls_config['metrics_recorder'] = get_mrec_from_config(config_dict.get('metrics_recorder'))
    cls_config['model'] = get_model_from_config(config_dict.get('model'))
    cls_config['fusion'] = get_fusion_from_config(config_dict.get('fusion'))
    cls_config['connection'] = get_connection_from_config(
        config_dict.get('connection'))
    cls_config['protocol_handler'] = get_ph_from_config(
        config_dict.get('protocol_handler'))
    cls_config['hyperparams'] = config_dict.get('hyperparams')
    cls_config['aggregator'] = config_dict.get('aggregator')
    cls_config['local_training'] = get_lt_from_config(
        config_dict.get('local_training'))
    cls_config['privacy'] = config_dict.get('privacy')
    cls_config['metrics'] = get_mh_from_config(config_dict.get('metrics'))
    cls_config['evidencia'] = get_evidencia_from_config(config_dict.get('evidencia'))
    cls_config['key_generators'] = config_dict.get('key_generators')
    cls_config['access_white_lst'] = config_dict.get('access_white_lst')

    return cls_config


def get_connection_from_config(config):
    """ Load connection class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    connection_config = {}
    if config:
        try:
            connection_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])
            connection_config['info'] = config['info']
            connection_config['sync'] = bool(config.get('sync', False))

        except InvalidConfigurationException as ex:
            logger.exception(ex)

            raise InvalidConfigurationException(
                'Error occurred while loading connection config.')

    else:
        logger.info('No connection config provided for this setup.')

    return connection_config


def get_data_from_config(config):
    """ Load data class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    data_config = None
    if config:
        data_config = {}
        try:
            if 'class' in config:
                data_config['cls_ref'] = config['class']
            else:
                data_config['cls_ref'] = get_class_by_name(
                    config['path'], config['name'])
            data_config['info'] = config['info']
        except InvalidConfigurationException as ex:
            logger.exception(ex)
            raise InvalidConfigurationException(
                'Error occurred while loading data config.')

    else:
        logger.info('No data config provided for this setup.')
    return data_config


def get_model_from_config(config):
    """ Load model class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    model_config = None
    if config:
        model_config = {}
        try:
            model_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])

            model_config['spec'] = config['spec']

            if 'info' in config:
                model_config['info'] = config['info']

            if 'model_file' in config:
                model_config['model_file'] = config['model_file']

        except Exception as ex:
            logger.exception(ex)
            raise InvalidConfigurationException(
                'Error occurred while loading model config.')

    else:
        logger.info('No model config provided for this setup.')

    return model_config


def get_fusion_from_config(config):
    """
    Load fusion class information from params provided in config file.

    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    fusion_config = {}
    if config:
        try:
            fusion_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])
            if 'info' in config:
                fusion_config['info'] = config['info']
        except Exception as ex:
            logger.exception(ex)

            raise InvalidConfigurationException(
                'Error occurred while loading fusion config.')

    else:
        logger.debug('No fusion config provided for this setup.')

    return fusion_config


def get_ph_from_config(config):
    """ Load ph class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    ph_config = {}
    if config:
        try:
            ph_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])
            if 'info' in config:
                ph_config['info'] = config['info']
        except Exception as ex:
            logger.exception(ex)

            raise InvalidConfigurationException(
                'Error occurred while loading protocol handler config.')

    else:
        logger.info('No ph config provided for this setup.')

    return ph_config


def get_evidencia_from_config(config):
    """ Evidencia recorder class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    evidencia_config = {}
    if config:
        try:
            evidencia_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])
            if 'info' in config:
                evidencia_config['info'] = config['info']

        except Exception as ex:
            logger.exception(ex)

            raise InvalidConfigurationException(
                'Error occurred while loading evidencia recorder config.')

    else:
        logger.info('No evidencia recordeer config provided for this setup.')

    return evidencia_config


def get_lt_from_config(config):
    """ Load local training class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    lt_config = {}
    if config:
        try:
            lt_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])
            if 'info' in config:
                lt_config['info'] = config['info']

        except Exception as ex:
            logger.exception(ex)

            raise InvalidConfigurationException(
                'Error occurred while loading local training config.')

    else:
        logger.debug('No local training config provided for this setup.')

    return lt_config


def get_mh_from_config(config):
    """ Load metrics handler class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    mh_config = None
    if config:
        mh_config = {}
        try:
            mh_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])
            mh_config['info'] = config.get('info')
        except InvalidConfigurationException as ex:
            logger.exception(ex)
            raise InvalidConfigurationException(
                'Error occurred while loading mh config.')

    else:
        logger.info('No metrics config provided for this setup.')
    return mh_config


def get_mrec_from_config(config):
    """ Load metrics recorder class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    # TODO: validation needs to be added
    pm_config = None
    if config:
        pm_config = {}
        try:
            pm_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])
            pm_config['output_file'] = config['output_file']
            pm_config['output_type'] = config['output_type']
            pm_config['compute_pre_train_eval'] = config['compute_pre_train_eval']
            pm_config['compute_post_train_eval'] = config['compute_post_train_eval']
        except InvalidConfigurationException as ex:
            logger.exception(ex)
            raise InvalidConfigurationException(
                'Error occurred while loading data config.')

    else:
        logger.info('No metrics recorder config provided for this setup.')
    return pm_config


def get_key_generator_from_config(config):
    """ Load data class information from params provided in config file
    :param config: dictionary of configuration
    :type config: `dict`
    :return: dictionary of class references
    :rtype: `dict`
    """
    key_generator_config = None
    if config:
        key_generator_config = {}
        try:
            key_generator_config['cls_ref'] = get_class_by_name(
                config['path'], config['name'])
            key_generator_config['sec_config'] = config.get('sec_config')
        except InvalidConfigurationException as ex:
            logger.exception(ex)
            raise InvalidConfigurationException(
                'Error occurred while loading mh config.')

    else:
        logger.info('No data config provided for this setup.')
    return key_generator_config


def get_class_by_name(path, name_class):
    """
    Gets object according to the name_class and the path where it is located.

    :param path: logical location of the object
    :type path: `str`
    :param name_class: name of the object
    :type name_class: `str`

    :return: Object
    """
    try:

        cls_ref = get_attr_from_path(path, name_class)
    except Exception as ex:
        logger.exception(ex)
        raise InvalidConfigurationException(
            'Error occurred while loading class '+name_class+'from path ' + path)
    return cls_ref


def get_attr_from_path(path, attr):
    """
    Gets attr from a file with given path.

    :param path: logical location of the object
    :type path: `str`
    :param attr: name of the attr
    :type attr: `str`

    :return: Object/method
    """
    file_path = Path(PurePath(path))
    if is_file_path(path):

        absolute_path = get_absolute_path(path)

        if absolute_path:
            loader = SourceFileLoader(file_path.stem, absolute_path)
            module = loader.load_module()
        else:
            raise InvalidConfigurationException(
                "File does not exist. "+path+" not found in current or working directory.")

    else:
        module = importlib.import_module(path)

    return getattr(module, attr)


def is_file_path(path):
    """
    Validate if a given path is directory file.

    :param path: path given in configuration
    :type path: `str`

    :return: Boolean
    """
    file_path = Path(PurePath(path))

    return file_path.suffix == '.py'


def get_absolute_path(path):
    """
    Get absolute path for `file path` given in config.
    If given path is relative- then absolute path is constructed by adding
    either the current directory or working directory to the path.
    If a file is found in current directory then working directory is not
    checked.

    :param path: logical location of the object
    :type path: `str`

    :return: absolute path
    """
    file_path = Path(PurePath(path))
    absolute_path = path  # when absolute path is provided

    if file_path.is_absolute == True and not file_path.exists():
        raise InvalidConfigurationException(
            "File does not exist. "+path+" not found.")

    elif get_current_dir_ap(path):
        absolute_path = get_current_dir_ap(path)

    else:
        absolute_path = get_working_dir_ap(path)

    if absolute_path is None or os.path.exists(absolute_path) is False:
        raise InvalidConfigurationException(
            "File does not exist. "+path+" not found in current or working directory.")

    return absolute_path


def get_current_dir_ap(path):
    """
    Get absolute path using current directory for `file path` given in config.
    If given path is relative, then absolute path is constructed by adding
    current directory as a prefix to path

    :param path: logical location of the object
    :type path: `str`

    :return: absolute path
    """
    file_path = Path(PurePath(path))

    current_dir = os.getcwd()
    absolute_path = Path(
        PurePath(current_dir).joinpath(PurePath(file_path)))
    if absolute_path.exists():
        return str(absolute_path)

    return None


def get_working_dir_ap(path):
    """
    Get absolute path using working_dir env variable for `file path` given in config.
    If given path is relative - then absolute path is constructed by adding
    working directory as a prefix to path

    :param path: logical location of the object
    :type path: `str`

    :return: absolute path
    """
    file_path = Path(PurePath(path))
    working_dir = fl_envs.working_directory
    absolute_path = Path(
        PurePath(working_dir).joinpath(PurePath(file_path)))
    if absolute_path.exists():
        return str(absolute_path)

    return None


def convert_zip_to_bytes(filename):
    """
    Open a zip file and return a byte array. 
    :param filename: name of the file
    :type filename: `str`
    :return: byte array
    """
    byte_arr = bytearray()

    if filename is not None:
        try:
            abs_path = get_absolute_path(filename)
            with open(abs_path, "rb") as f:
                byte_arr = f.read()

        except InvalidConfigurationException as ex:
            logger.error(str(ex))
            raise InvalidConfigurationException("Zip file does not exist.")
        except Exception as identifier:
            raise FLException(
                "Error while converting zip file to bytes.")

    return byte_arr


def convert_bytes_to_zip(byte_arr, destination_file):
    """
    Copy the bytes into a file 
    :param byte_arr: byte array which needs to be copied into destination file
    :param destination_file: name of the file
    :type destination_file: `str`
    :return: byte array
    """
    try:
        f_out = open(destination_file, 'w+b')
        f_out.write(byte_arr)
        f_out.close

    except Exception as identifier:
        raise FLException(
            "Error while creating a file at " + destination_file)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from DAG import DAG
from IBMSparkPipeline import IBMSparkPipeline
from IBMSparkPipelineModel import IBMSparkPipelineModel
from Source import Source
from Sink import Sink

# Set the version of module
__version__ =  '0.1.457'



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from typing import TYPE_CHECKING, Any, Dict, Union, List, Optional

import pandas as pd
from pandas import DataFrame

from .base_deployment import BaseDeployment
from ..wml_client_error import WMLClientError

if TYPE_CHECKING:
    from sklearn.pipeline import Pipeline
    from pandas import DataFrame
    from numpy import ndarray
    from ..workspace import WorkSpace

__all__ = [
    "WebService"
]


class WebService(BaseDeployment):
    """An Online Deployment class aka. WebService.
    With this class object you can manage any online (WebService) deployment.

    :param source_wml_credentials: credentials to Watson Machine Learning instance where training was performed
    :type source_wml_credentials: dict

    :param source_project_id: ID of the Watson Studio project where training was performed
    :type source_project_id: str, optional

    :param source_space_id: ID of the Watson Studio Space where training was performed
    :type source_space_id: str, optional

    :param target_wml_credentials: credentials to Watson Machine Learning instance where you want to deploy
    :type target_wml_credentials: dict

    :param target_project_id: ID of the Watson Studio project where you want to deploy
    :type target_project_id: str, optional

    :param target_space_id: ID of the Watson Studio Space where you want to deploy
    :type target_space_id: str, optional
    """

    def __init__(self,
                 source_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 source_project_id: str = None,
                 source_space_id: str = None,
                 target_wml_credentials: Union[dict, 'WorkSpace'] = None,
                 target_project_id: str = None,
                 target_space_id: str = None,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None):

        super().__init__(
            deployment_type='online',
            source_wml_credentials=source_wml_credentials,
            source_project_id=source_project_id,
            source_space_id=source_space_id,
            target_wml_credentials=target_wml_credentials,
            target_project_id=target_project_id,
            target_space_id=target_space_id,
            wml_credentials=wml_credentials,
            project_id=project_id,
            space_id=space_id
        )

        self.name = None
        self.scoring_url = None
        self.id = None
        self.asset_id = None

    def __repr__(self):
        return f"name: {self.name}, id: {self.id}, scoring_url: {self.scoring_url}, asset_id: {self.asset_id}"

    def __str__(self):
        return f"name: {self.name}, id: {self.id}, scoring_url: {self.scoring_url}, asset_id: {self.asset_id}"

    def create(self,
               model: str,
               deployment_name: str,
               serving_name: str = None,
               metadata: Optional[Dict] = None,
               training_data: Optional[Union['DataFrame', 'ndarray']] = None,
               training_target: Optional[Union['DataFrame', 'ndarray']] = None,
               experiment_run_id: Optional[str] = None,
               hardware_spec: Optional[dict] = None) -> None:
        """Create deployment from a model.

        :param model: AutoAI model name
        :type model: str

        :param deployment_name: name of the deployment
        :type deployment_name: str

        :param training_data: training data for the model
        :type training_data: pandas.DataFrame or numpy.ndarray, optional

        :param training_target: target/label data for the model
        :type training_target: pandas.DataFrame or numpy.ndarray, optional

        :param serving_name: serving name of the deployment
        :type serving_name: str, optional

        :param metadata: model meta properties
        :type metadata: dict, optional

        :param experiment_run_id: ID of a training/experiment (only applicable for AutoAI deployments)
        :type experiment_run_id: str, optional

        :param hardware_spec: hardware specification for deployment
        :type hardware_spec: dict, optional

        **Example**

        .. code-block:: python

            from ibm_watson_machine_learning.deployment import WebService

            deployment = WebService(
                   wml_credentials={
                         "apikey": "...",
                         "iam_apikey_description": "...",
                         "iam_apikey_name": "...",
                         "iam_role_crn": "...",
                         "iam_serviceid_crn": "...",
                         "instance_id": "...",
                         "url": "https://us-south.ml.cloud.ibm.com"
                       },
                    project_id="...",
                    space_id="...")

            deployment.create(
                   experiment_run_id="...",
                   model=model,
                   deployment_name='My new deployment',
                   serving_name='my_new_deployment'
               )
        """
        return super().create(model=model,
                              deployment_name=deployment_name,
                              metadata=metadata,
                              serving_name=serving_name,
                              training_data=training_data,
                              training_target=training_target,
                              experiment_run_id=experiment_run_id,
                              deployment_type='online',
                              hardware_spec=hardware_spec)

    @BaseDeployment._project_to_space_to_project
    def get_params(self) -> Dict:
        """Get deployment parameters."""
        return super().get_params()

    @BaseDeployment._project_to_space_to_project
    def score(self, payload: Union[dict, 'DataFrame'] = pd.DataFrame(), transaction_id: str = None) -> Dict[str, List]:
        """Online scoring on WML. Payload is passed to the WML scoring endpoint where model have been deployed.

        :param payload: DataFrame with data to test the model or dictionary with keys `observations`
            and `supporting_features` and DataFrames with data for `observations` and `supporting_features`
            to score forecasting models
        :type payload: pandas.DataFrame or dict

        :param transaction_id: can be used to indicate under which id the records will be saved into payload table
            in IBM OpenScale
        :type transaction_id: str, optional

        :return: dictionary with list od model output/predicted targets
        :rtype: dict

        **Examples**

        .. code-block:: python

            predictions = web_service.score(payload=test_data)
            print(predictions)

            # Result:
            # {'predictions':
            #     [{
            #         'fields': ['prediction', 'probability'],
            #         'values': [['no', [0.9221385608558003, 0.07786143914419975]],
            #                   ['no', [0.9798324002736079, 0.020167599726392187]]
            #     }]}

            predictions = web_service.score(payload={'observations': new_observations_df})
            predictions = web_service.score(payload={'observations': new_observations_df, 'supporting_features': supporting_features_df}) # supporting features time series forecasting sceanrio
        """
        return super().score(payload=payload, transaction_id=transaction_id)

    @BaseDeployment._project_to_space_to_project
    def delete(self, deployment_id: str = None) -> None:
        """Delete deployment on WML.

        :param deployment_id: ID of the deployment to delete, if empty, current deployment will be deleted
        :type deployment_id: str, optional

        **Example**

        .. code-block:: python

            deployment = WebService(workspace=...)
            # Delete current deployment
            deployment.delete()
            # Or delete a specific deployment
            deployment.delete(deployment_id='...')
        """
        super().delete(deployment_id=deployment_id, deployment_type='online')

    @BaseDeployment._project_to_space_to_project
    def list(self, limit=None) -> 'DataFrame':
        """List WML deployments.

        :param limit: set the limit of how many deployments to list,
            default is `None` (all deployments should be fetched)
        :type limit: int, optional

        :return: Pandas DataFrame with information about deployments
        :rtype: pandas.DataFrame

        **Example**

        .. code-block:: python

            deployment = WebService(workspace=...)
            deployments_list = deployment.list()
            print(deployments_list)

            # Result:
            #                  created_at  ...  status
            # 0  2020-03-06T10:50:49.401Z  ...   ready
            # 1  2020-03-06T13:16:09.789Z  ...   ready
            # 4  2020-03-11T14:46:36.035Z  ...  failed
            # 3  2020-03-11T14:49:55.052Z  ...  failed
            # 2  2020-03-11T15:13:53.708Z  ...   ready
        """
        return super().list(limit=limit, deployment_type='online')

    @BaseDeployment._project_to_space_to_project
    def get(self, deployment_id: str) -> None:
        """Get WML deployment.

        :param deployment_id: ID of the deployment to work with
        :type deployment_id: str

        **Example**

        .. code-block:: python

            deployment = WebService(workspace=...)
            deployment.get(deployment_id="...")
        """
        super().get(deployment_id=deployment_id, deployment_type='online')

    @BaseDeployment._project_to_space_to_project
    def _deploy(self,
                pipeline_model: 'Pipeline',
                deployment_name: str,
                meta_props: Dict,
                serving_name=None,
                result_client=None,
                hardware_spec=None) -> Dict:
        """Deploy model into WML.

        :param pipeline_model: model of the pipeline to deploy
        :type pipeline_model: Pipeline or str

        :param deployment_name: name of the deployment
        :type deployment_name: str

        :param meta_props: model meta properties
        :type meta_props: dict

        :param serving_name: serving name of the deployment
        :type serving_name: str

        :param result_client: tuple with Result DataConnection object and initialized COS client
        :rtype: tuple[DataConnection, resource]

        :return: deployment details
        :rtype: dict
        """
        asset_uid = self._publish_model(pipeline_model=pipeline_model,
                                        meta_props=meta_props)

        self.asset_id = asset_uid

        conf_names = self._target_workspace.wml_client.deployments.ConfigurationMetaNames

        deployment_props = {
            conf_names.NAME: deployment_name,
            conf_names.ONLINE: {}
        }

        if hardware_spec:
            deployment_props[conf_names.HARDWARE_SPEC] = hardware_spec

        if serving_name:
            deployment_props[conf_names.ONLINE]["parameters"] = {conf_names.SERVING_NAME: serving_name}

        print("Deploying model {} using V4 client.".format(asset_uid))
        try:

            deployment_details = self._target_workspace.wml_client.deployments.create(
                artifact_uid=asset_uid,
                meta_props=deployment_props)
            self.deployment_id = self._target_workspace.wml_client.deployments.get_uid(deployment_details)

        except WMLClientError as e:
            raise e

        return deployment_details
